import sys
import typing as t
from collections import abc
from datetime import date

import pydantic
import pytest
from django.core.exceptions import FieldError, ValidationError
from django.db import models
from django.db.migrations.writer import MigrationWriter
from django_pydantic_field import fields

from .conftest import InnerSchema, SampleDataclass


class SampleModel(models.Model):
    sample_field: InnerSchema = fields.SchemaField(config={"frozen": True, "allow_mutation": False})
    sample_list: t.List[InnerSchema] = fields.SchemaField()
    sample_seq: t.Sequence[InnerSchema] = fields.SchemaField(schema=t.List[InnerSchema], default=list)

    class Meta:
        app_label = "test_app"


class SampleForwardRefModel(models.Model):
    annotated_field: "SampleSchema" = fields.SchemaField()
    field = fields.SchemaField(schema=t.ForwardRef("SampleSchema"))

    class Meta:
        app_label = "test_app"


class SampleSchema(pydantic.BaseModel):
    field: int = 1


def test_sample_field():
    sample_field = fields.PydanticSchemaField(schema=InnerSchema)
    existing_instance = InnerSchema(stub_str="abc", stub_list=[date(2022, 7, 1)])
    expected_encoded = '{"stub_str": "abc", "stub_int": 1, "stub_list": ["2022-07-01"]}'

    assert sample_field.get_prep_value(existing_instance) == expected_encoded
    assert sample_field.to_python(expected_encoded) == existing_instance


def test_sample_field_with_raw_data():
    sample_field = fields.PydanticSchemaField(schema=InnerSchema)
    existing_raw = {"stub_str": "abc", "stub_list": [date(2022, 7, 1)]}
    expected_encoded = '{"stub_str": "abc", "stub_int": 1, "stub_list": ["2022-07-01"]}'

    assert sample_field.get_prep_value(existing_raw) == expected_encoded
    assert sample_field.to_python(expected_encoded) == InnerSchema(**existing_raw)


def test_simple_model_field():
    sample_field = SampleModel._meta.get_field("sample_field")
    assert sample_field.schema == InnerSchema

    sample_list_field = SampleModel._meta.get_field("sample_list")
    assert sample_list_field.schema == t.List[InnerSchema]

    sample_seq_field = SampleModel._meta.get_field("sample_seq")
    assert sample_seq_field.schema == t.List[InnerSchema]

    existing_raw_field = {"stub_str": "abc", "stub_list": [date(2022, 7, 1)]}
    existing_raw_list = [{"stub_str": "abc", "stub_list": []}]

    instance = SampleModel(sample_field=existing_raw_field, sample_list=existing_raw_list)

    expected_instance = InnerSchema(stub_str="abc", stub_list=[date(2022, 7, 1)])
    expected_list = [InnerSchema(stub_str="abc", stub_list=[])]

    assert instance.sample_field == expected_instance
    assert instance.sample_list == expected_list


def test_null_field():
    field = fields.SchemaField(InnerSchema, null=True, default=None)
    assert field.to_python(None) is None
    assert field.get_prep_value(None) is None

    field = fields.SchemaField(t.Optional[InnerSchema], null=True, default=None)
    assert field.get_prep_value(None) is None


def test_untyped_model_field_raises():
    with pytest.raises(FieldError):
        class UntypedModel(models.Model):
            sample_field = fields.SchemaField()

            class Meta:
                app_label = "test_app"


def test_forwardrefs_deferred_resolution():
    obj = SampleForwardRefModel(field={}, annotated_field={})
    assert isinstance(obj.field, SampleSchema)
    assert isinstance(obj.annotated_field, SampleSchema)


@pytest.mark.parametrize("forward_ref", [
    "InnerSchema",
    t.ForwardRef("SampleDataclass"),
    t.List["int"]
])
def test_resolved_forwardrefs(forward_ref):
    class ModelWithForwardRefs(models.Model):
        field: forward_ref = fields.SchemaField()

        class Meta:
            app_label = "test_app"


@pytest.mark.parametrize("field", [
    fields.PydanticSchemaField(schema=InnerSchema, default=InnerSchema(stub_str="abc", stub_list=[date(2022, 7, 1)])),
    fields.PydanticSchemaField(schema=InnerSchema, default=(("stub_str", "abc"), ("stub_list", [date(2022, 7, 1)]))),
    fields.PydanticSchemaField(schema=InnerSchema, default={"stub_str": "abc", "stub_list": [date(2022, 7, 1)]}),
    fields.PydanticSchemaField(schema=InnerSchema, null=True, default=None),
    fields.PydanticSchemaField(schema=SampleDataclass, default={"stub_str": "abc", "stub_list": [date(2022, 7, 1)]}),
    fields.PydanticSchemaField(schema=t.Optional[InnerSchema], null=True, default=None),
])
def test_field_serialization(field):
    _test_field_serialization(field)


@pytest.mark.skipif(sys.version_info < (3, 9), reason="Built-in type subscription supports only in 3.9+")
@pytest.mark.parametrize("field_factory", [
    lambda: fields.PydanticSchemaField(schema=list[InnerSchema], default=list),
    lambda: fields.PydanticSchemaField(schema=dict[str, InnerSchema], default=list),
    lambda: fields.PydanticSchemaField(schema=abc.Sequence[InnerSchema], default=list),
    lambda: fields.PydanticSchemaField(schema=abc.Mapping[str, InnerSchema], default=dict),
])
def test_field_builtin_annotations_serialization(field_factory):
    _test_field_serialization(field_factory())


@pytest.mark.skipif(sys.version_info < (3, 10), reason="Union type syntax supported only in 3.10+")
def test_field_union_type_serialization():
    field = fields.PydanticSchemaField(schema=(InnerSchema | None), null=True, default=None)
    _test_field_serialization(field)


@pytest.mark.skipif(sys.version_info >= (3, 9), reason="Should test against builtin generic types")
@pytest.mark.parametrize("field", [
    fields.PydanticSchemaField(schema=t.List[InnerSchema], default=list),
    fields.PydanticSchemaField(schema=t.Dict[str, InnerSchema], default=list),
    fields.PydanticSchemaField(schema=t.Sequence[InnerSchema], default=list),
    fields.PydanticSchemaField(schema=t.Mapping[str, InnerSchema], default=dict),
])
def test_field_typing_annotations_serialization(field):
    _test_field_serialization(field)


@pytest.mark.skipif(sys.version_info < (3, 9), reason="Typing-to-builtin migrations is reasonable only on py >= 3.9")
@pytest.mark.parametrize("old_field, new_field", [
    (
        lambda: fields.PydanticSchemaField(schema=t.List[InnerSchema], default=list),
        lambda: fields.PydanticSchemaField(schema=list[InnerSchema], default=list),
    ), (
        lambda: fields.PydanticSchemaField(schema=t.Dict[str, InnerSchema], default=list),
        lambda: fields.PydanticSchemaField(schema=dict[str, InnerSchema], default=list),
    ), (
        lambda: fields.PydanticSchemaField(schema=t.Sequence[InnerSchema], default=list),
        lambda: fields.PydanticSchemaField(schema=abc.Sequence[InnerSchema], default=list),
    ), (
        lambda: fields.PydanticSchemaField(schema=t.Mapping[str, InnerSchema], default=dict),
        lambda: fields.PydanticSchemaField(schema=abc.Mapping[str, InnerSchema], default=dict),
    ), (
        lambda: fields.PydanticSchemaField(schema=t.Mapping[str, InnerSchema], default=dict),
        lambda: fields.PydanticSchemaField(schema=abc.Mapping[str, InnerSchema], default=dict),
    )
])
def test_field_typing_to_builtin_serialization(old_field, new_field):
    old_field, new_field = old_field(), new_field()

    _, _, args, kwargs = old_field.deconstruct()

    reconstructed_field = fields.PydanticSchemaField(*args, **kwargs)
    assert old_field.get_default() == new_field.get_default() == reconstructed_field.get_default()
    assert new_field.schema == reconstructed_field.schema

    deserialized_field = reconstruct_field(serialize_field(old_field))
    assert old_field.get_default() == deserialized_field.get_default() == new_field.get_default()
    assert new_field.schema == deserialized_field.schema


@pytest.mark.parametrize("field, flawed_data", [
    (fields.PydanticSchemaField(schema=InnerSchema), {}),
    (fields.PydanticSchemaField(schema=t.List[InnerSchema]), [{}]),
    (fields.PydanticSchemaField(schema=t.Dict[int, float]), {"1": "abc"}),
])
def test_field_validation_exceptions(field, flawed_data):
    with pytest.raises(ValidationError):
        field.to_python(flawed_data)

def test_model_validation_exceptions():
    with pytest.raises(ValidationError):
        SampleModel(sample_field=1)
    with pytest.raises(ValidationError):
        SampleModel(sample_field={"stub_list": {}, "stub_str": ""})

    valid_initial = SampleModel(
        sample_field={"stub_list": [], "stub_str": ""},
        sample_list=[],
        sample_seq=[],
    )
    with pytest.raises(ValidationError):
        valid_initial.sample_field = 1


@pytest.mark.parametrize("export_kwargs", [
    {"include": {"stub_str", "stub_int"}},
    {"exclude": {"stub_list"}},
    {"by_alias": True},
    {"exclude_unset": True},
    {"exclude_defaults": True},
    {"exclude_none": True}
])
def test_export_kwargs_support(export_kwargs):
    field = fields.PydanticSchemaField(
        schema=InnerSchema,
        default=InnerSchema(stub_str="", stub_list=[]),
        **export_kwargs,
    )
    _test_field_serialization(field)

    existing_instance = InnerSchema(stub_str="abc", stub_list=[date(2022, 7, 1)])
    assert field.get_prep_value(existing_instance)


def _test_field_serialization(field):
    _, _, args, kwargs = field.deconstruct()

    reconstructed_field = fields.PydanticSchemaField(*args, **kwargs)
    assert field.get_default() == reconstructed_field.get_default()
    assert field.schema == reconstructed_field.schema

    deserialized_field = reconstruct_field(serialize_field(field))
    assert deserialized_field.get_default() == field.get_default()
    assert field.schema == deserialized_field.schema


def serialize_field(field: fields.PydanticSchemaField) -> str:
    serialized_field, _ = MigrationWriter.serialize(field)
    return serialized_field


def reconstruct_field(field_repr: str) -> fields.PydanticSchemaField:
    return eval(field_repr, globals(), sys.modules)



import typing as t

try:
    from typing import get_args
except ImportError:
    from typing_extensions import get_args

from django.conf import settings
from pydantic import BaseModel, ValidationError

from rest_framework import exceptions, parsers, renderers, serializers
from rest_framework.schemas import openapi
from rest_framework.schemas.utils import is_list_view

from . import base

__all__ = (
    "SchemaField",
    "SchemaRenderer",
    "SchemaParser",
    "AutoSchema",
)

if t.TYPE_CHECKING:
    RequestResponseContext = t.Mapping[str, t.Any]


class AnnotatedSchemaT(t.Generic[base.ST]):
    schema_ctx_attr: t.ClassVar[str] = "schema"
    require_explicit_schema: t.ClassVar[bool] = False
    _cached_annotation_schema: t.Type[BaseModel]

    def get_schema(self, ctx: "RequestResponseContext") -> t.Optional[t.Type[BaseModel]]:
        schema = self.get_context_schema(ctx)
        if schema is None:
            schema = self.get_annotation_schema(ctx)

        if self.require_explicit_schema and schema is None:
            raise ValueError(
                "Schema should be either explicitly set with annotation "
                "or passed in the context"
            )

        return schema

    def get_context_schema(self, ctx: "RequestResponseContext"):
        schema = ctx.get(self.schema_ctx_attr)
        if schema is not None:
            schema = base.wrap_schema(schema)
            base.prepare_schema(schema, ctx.get("view"))

        return schema

    def get_annotation_schema(self, ctx: "RequestResponseContext"):
        try:
            schema = self._cached_annotation_schema
        except AttributeError:
            try:
                schema = get_args(self.__orig_class__)[0]  # type: ignore
            except (AttributeError, IndexError):
                return None

            self._cached_annotation_schema = schema = base.wrap_schema(schema)
            base.prepare_schema(schema, ctx.get("view"))

        return schema


class SchemaField(serializers.Field, t.Generic[base.ST]):
    decoder: "base.SchemaDecoder[base.ST]"
    _is_prepared_schema: bool = False

    def __init__(
        self,
        schema: t.Type["base.ST"],
        config: t.Optional["base.ConfigType"] = None,
        **kwargs,
    ):
        nullable = kwargs.get("allow_null", False)

        self.schema = field_schema = base.wrap_schema(schema, config, nullable)
        self.export_params = base.extract_export_kwargs(kwargs, dict.pop)
        self.decoder = base.SchemaDecoder(field_schema)
        super().__init__(**kwargs)

    def bind(self, field_name, parent):
        if not self._is_prepared_schema:
            base.prepare_schema(self.schema, parent)
            self._is_prepared_schema = True

        super().bind(field_name, parent)

    def to_internal_value(self, data: t.Any) -> t.Optional["base.ST"]:
        try:
            return self.decoder.decode(data)
        except ValidationError as e:
            raise serializers.ValidationError(e.errors(), self.field_name)

    def to_representation(self, value: t.Optional["base.ST"]) -> t.Any:
        obj = self.schema.parse_obj(value)
        raw_obj = obj.dict(**self.export_params)
        return raw_obj["__root__"]


class SchemaRenderer(AnnotatedSchemaT[base.ST], renderers.JSONRenderer):
    schema_ctx_attr = "render_schema"

    def render(self, data, accepted_media_type=None, renderer_context=None):
        renderer_context = renderer_context or {}
        response = renderer_context.get("response")
        if response is not None and response.exception:
            return super().render(data, accepted_media_type, renderer_context)

        try:
            json_str = self.render_data(data, renderer_context)
        except ValidationError as e:
            json_str = e.json().encode()
        except AttributeError:
            json_str = super().render(data, accepted_media_type, renderer_context)

        return json_str

    def render_data(self, data, renderer_ctx) -> bytes:
        schema = self.get_schema(renderer_ctx or {})
        if schema is not None:
            data = schema(__root__=data)

        export_kw = base.extract_export_kwargs(renderer_ctx)
        json_str = data.json(**export_kw, ensure_ascii=self.ensure_ascii)
        return json_str.encode()


class SchemaParser(AnnotatedSchemaT[base.ST], parsers.JSONParser):
    schema_ctx_attr = "parser_schema"
    renderer_class = SchemaRenderer
    require_explicit_schema = True

    def parse(self, stream, media_type=None, parser_context=None):
        parser_context = parser_context or {}
        encoding = parser_context.get("encoding", settings.DEFAULT_CHARSET)
        schema = t.cast(BaseModel, self.get_schema(parser_context))

        try:
            return schema.parse_raw(stream.read(), encoding=encoding).__root__
        except ValidationError as e:
            raise exceptions.ParseError(e.errors())


class AutoSchema(openapi.AutoSchema):
    get_request_serializer: t.Callable
    _get_reference: t.Callable

    def map_field(self, field: serializers.Field):
        if isinstance(field, SchemaField):
            return field.schema.schema()
        return super().map_field(field)

    def map_parsers(self, path: str, method: str):
        request_types: t.List[t.Any] = []
        parser_ctx = self.view.get_parser_context(None)

        for parser_type in self.view.parser_classes:
            parser = parser_type()

            if isinstance(parser, SchemaParser):
                schema = self._extract_openapi_schema(parser, parser_ctx)
                if schema is not None:
                    request_types.append((parser.media_type, schema))
                else:
                    request_types.append(parser.media_type)
            else:
                request_types.append(parser.media_type)

        return request_types

    def map_renderers(self, path: str, method: str):
        response_types: t.List[t.Any] = []
        renderer_ctx = self.view.get_renderer_context()

        for renderer_type in self.view.renderer_classes:
            renderer = renderer_type()

            if isinstance(renderer, SchemaRenderer):
                schema = self._extract_openapi_schema(renderer, renderer_ctx)
                if schema is not None:
                    response_types.append((renderer.media_type, schema))
                else:
                    response_types.append(renderer.media_type)

            elif not isinstance(renderer, renderers.BrowsableAPIRenderer):
                response_types.append(renderer.media_type)

        return response_types

    def get_request_body(self, path: str, method: str):
        if method not in ('PUT', 'PATCH', 'POST'):
            return {}

        self.request_media_types = self.map_parsers(path, method)
        serializer = self.get_request_serializer(path, method)
        content_schemas = {}

        for request_type in self.request_media_types:
            if isinstance(request_type, tuple):
                media_type, request_schema = request_type
                content_schemas[media_type] = {"schema": request_schema}
            else:
                serializer_ref =  self._get_reference(serializer)
                content_schemas[request_type] = {"schema": serializer_ref}

        return {'content': content_schemas}

    def get_responses(self, path: str, method: str):
        if method == "DELETE":
            return {"204": {"description": ""}}

        self.response_media_types = self.map_renderers(path, method)
        status_code = "201" if method == "POST" else "200"
        content_types = {}

        for response_type in self.response_media_types:
            if isinstance(response_type, tuple):
                media_type, response_schema = response_type
                content_types[media_type] = {"schema": response_schema}
            else:
                response_schema = self._get_serializer_response_schema(path, method)
                content_types[response_type] = {"schema": response_schema}

        return {
            status_code: {
                "content": content_types,
                "description": "",
            }
        }

    def _extract_openapi_schema(self, schemable: AnnotatedSchemaT, ctx: "RequestResponseContext"):
        schema_model = schemable.get_schema(ctx)
        if schema_model is not None:
            return schema_model.schema()
        return None

    def _get_serializer_response_schema(self, path, method):
        serializer = self.get_response_serializer(path, method)

        if not isinstance(serializer, serializers.Serializer):
            item_schema = {}
        else:
            item_schema = self._get_reference(serializer)

        if is_list_view(path, method, self.view):
            response_schema = {
                "type": "array",
                "items": item_schema,
            }
            paginator = self.get_paginator()
            if paginator:
                response_schema = paginator.get_paginated_response_schema(response_schema)
        else:
            response_schema = item_schema
        return response_schema



import typing as t

import pydantic
from django.core.serializers.json import DjangoJSONEncoder
from pydantic.config import get_config, inherit_config
from pydantic.typing import display_as_type

from .utils import get_local_namespace

__all__ = (
    "SchemaEncoder",
    "SchemaDecoder",
    "wrap_schema",
    "prepare_schema",
    "extract_export_kwargs",
)

ST = t.TypeVar("ST", bound="SchemaT")

if t.TYPE_CHECKING:
    from pydantic.dataclasses import DataclassClassOrWrapper

    SchemaT = t.Union[
        pydantic.BaseModel,
        DataclassClassOrWrapper,
        t.Sequence[t.Any],
        t.Mapping[str, t.Any],
        t.Set[t.Any],
        t.FrozenSet[t.Any],
    ]

    ModelType = t.Type[pydantic.BaseModel]
    ConfigType = t.Union[pydantic.ConfigDict, t.Type[pydantic.BaseConfig], t.Type]


class SchemaEncoder(DjangoJSONEncoder):
    def __init__(
        self,
        *args,
        schema: "ModelType",
        export=None,
        raise_errors: bool = False,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.schema = schema
        self.export_params = export or {}
        self.raise_errors = raise_errors

    def encode(self, obj):
        try:
            data = self.schema(__root__=obj).json(**self.export_params)
        except pydantic.ValidationError:
            if self.raise_errors:
                raise

            # This branch used for expressions like .filter(data__contains={}).
            # We don't want that {} to be parsed as a schema.
            data = super().encode(obj)

        return data


class SchemaDecoder(t.Generic[ST]):
    def __init__(self, schema: "ModelType"):
        self.schema = schema

    def decode(self, obj: t.Any) -> "ST":
        if isinstance(obj, (str, bytes)):
            value = self.schema.parse_raw(obj).__root__  # type: ignore
        else:
            value = self.schema.parse_obj(obj).__root__  # type: ignore
        return value


def wrap_schema(
    schema: t.Union[t.Type["ST"], t.ForwardRef],
    config: t.Optional["ConfigType"] = None,
    allow_null: bool = False,
    **kwargs,
) -> "ModelType":
    type_name = _get_field_schema_name(schema)
    params = _get_field_schema_params(schema, config, allow_null, **kwargs)
    return pydantic.create_model(type_name, **params)


def prepare_schema(schema: "ModelType", owner: t.Any = None) -> None:
    namespace = get_local_namespace(owner)
    schema.update_forward_refs(**namespace)


def extract_export_kwargs(ctx: dict, extractor=dict.get) -> t.Dict[str, t.Any]:
    export_ctx = dict(
        exclude_defaults=extractor(ctx, "exclude_defaults", None),
        exclude_none=extractor(ctx, "exclude_none", None),
        exclude_unset=extractor(ctx, "exclude_unset", None),
        by_alias=extractor(ctx, "by_alias", None),
    )
    include_fields = extractor(ctx, "include", None)
    if include_fields is not None:
        export_ctx["include"] = {"__root__": include_fields}

    exclude_fields = extractor(ctx, "exclude", None)
    if exclude_fields is not None:
        export_ctx["exclude"] = {"__root__": exclude_fields}

    return {k: v for k, v in export_ctx.items() if v is not None}


def _get_field_schema_name(schema) -> str:
    return f"FieldSchema[{display_as_type(schema)}]"


def _get_field_schema_params(schema, config=None, allow_null=False, **kwargs) -> dict:
    root_model = t.Optional[schema] if allow_null else schema
    params: t.Dict[str, t.Any] = dict(kwargs, __root__=(root_model, ...))
    parent_config = getattr(schema, "Config", None)

    if config is not None:
        config = get_config(config)
        if parent_config is not None:
            config = inherit_config(config, parent_config)
    else:
        config = parent_config

    params["__config__"] = config
    return params



import typing as t

import django
import pydantic
import pytest
from django.core.exceptions import ValidationError
from django.db import models
from django.forms import Form, modelform_factory
from django_pydantic_field import fields, forms

from .conftest import InnerSchema


class SampleForwardRefFieldModel(models.Model):
    annotated_field: t.Optional["SampleSchema"] = fields.SchemaField(null=True)

    class Meta:
        app_label = "test_app"


class SampleForm(Form):
    field = forms.SchemaField(t.ForwardRef("SampleSchema"))


class SampleSchema(pydantic.BaseModel):
    field: int = 1


def test_form_schema_field():
    field = forms.SchemaField(InnerSchema)

    cleaned_data = field.clean('{"stub_str": "abc", "stub_list": ["1970-01-01"]}')
    assert cleaned_data == InnerSchema.parse_obj({"stub_str": "abc", "stub_list": ["1970-01-01"]})

def test_empty_form_values():
    field = forms.SchemaField(InnerSchema, required=False)
    assert field.clean("") is None
    assert field.clean(None) is None


def test_invalid_raises():
    field = forms.SchemaField(InnerSchema)
    with pytest.raises(ValidationError) as e:
        field.clean("")

    assert e.match("This field is required")

    with pytest.raises(ValidationError) as e:
        field.clean('{"stub_list": "abc"}')

    assert e.match("stub_str")
    assert e.match("stub_list")


@pytest.mark.xfail(
    django.VERSION[:2] < (4, 0),
    reason="Django < 4 has it's own feeling on bound fields resolution",
)
def test_forwardref_field():
    form = SampleForm(data={"field": '{"field": "2"}'})
    assert form.is_valid()


def test_model_formfield():
    field = fields.PydanticSchemaField(schema=InnerSchema)
    assert isinstance(field.formfield(), forms.SchemaField)


def test_forwardref_model_formfield():
    form_cls = modelform_factory(SampleForwardRefFieldModel, exclude=())
    form = form_cls(data={"annotated_field": '{"field": "2"}'})

    assert form.is_valid()
    cleaned_data = form.cleaned_data

    assert cleaned_data is not None
    assert cleaned_data["annotated_field"] == SampleSchema(field=2)


@pytest.mark.parametrize("export_kwargs", [
    {"include": {"stub_str", "stub_int"}},
    {"exclude": {"stub_list"}},
    {"exclude_unset": True},
    {"exclude_defaults": True},
    {"exclude_none": True},
    {"by_alias": True},
])
def test_form_field_export_kwargs(export_kwargs):
    field = forms.SchemaField(InnerSchema, required=False, **export_kwargs)
    value = InnerSchema.parse_obj({"stub_str": "abc", "stub_list": ["1970-01-01"]})
    assert field.prepare_value(value)



"""
Django Migration serializer helpers

[Built-in generic annotations](https://peps.python.org/pep-0585/)
    introduced in Python 3.9 are having a different semantics from `typing` collections.
    Due to how Django treats field serialization/reconstruction while writing migrations,
    it is not possible to distnguish between `types.GenericAlias` and any other regular types,
    thus annotations are being erased by `MigrationWriter` serializers.

    To mitigate this, I had to introduce custom container for schema deconstruction.

[Union types syntax](https://peps.python.org/pep-0604/)
    `typing.Union` and its special forms, like `typing.Optional`, have its own inheritance chain.
    Moreover, `types.UnionType`, introduced in 3.10, do not allow explicit type construction,
    only with `X | Y` syntax. Both cases require a dedicated serializer for migration writes.
"""
import sys
import types
import typing as t

try:
    from typing import get_args, get_origin
except ImportError:
    from typing_extensions import get_args, get_origin

from django.db.migrations.serializer import BaseSerializer, serializer_factory
from django.db.migrations.writer import MigrationWriter


class GenericContainer:
    __slots__ = "origin", "args"

    def __init__(self, origin, args=()):
        self.origin = origin
        self.args = args

    @classmethod
    def from_generic(cls, type_alias):
        return cls(get_origin(type_alias), get_args(type_alias))

    def reconstruct_type(self):
        if not self.args:
            return self.origin
        try:
            return self.origin[self.args]
        except TypeError:
            return GenericAlias(self.origin, self.args)

    def __repr__(self):
        return repr(self.reconstruct_type())

    __str__ = __repr__

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.origin, self.args == other.origin, other.args
        if isinstance(other, GenericTypes):
            return self == self.from_generic(other)
        return NotImplemented


class GenericSerializer(BaseSerializer):
    value: GenericContainer

    def serialize(self):
        value = self.value

        tp_repr, imports = serializer_factory(type(value)).serialize()
        orig_repr, orig_imports = serializer_factory(value.origin).serialize()
        imports.update(orig_imports)

        args = []
        for arg in value.args:
            arg_repr, arg_imports = serializer_factory(arg).serialize()
            args.append(arg_repr)
            imports.update(arg_imports)

        if args:
            args_repr = ", ".join(args)
            generic_repr = "%s(%s, (%s,))" % (tp_repr, orig_repr, args_repr)
        else:
            generic_repr = "%s(%s)" % (tp_repr, orig_repr)

        return generic_repr, imports


class TypingSerializer(BaseSerializer):
    def serialize(self):
        orig_module = self.value.__module__
        orig_repr = repr(self.value)

        if not orig_repr.startswith(orig_module):
            orig_repr = f"{orig_module}.{orig_repr}"

        return orig_repr, {f"import {orig_module}"}


if sys.version_info >= (3, 9):
    GenericAlias = types.GenericAlias
    GenericTypes: t.Tuple[t.Any, ...] = (
        GenericAlias,
        type(t.List[int]),
        type(t.List),
    )
else:
    # types.GenericAlias is missing, meaning python version < 3.9,
    # which has a different inheritance models for typed generics
    GenericAlias = type(t.List[int])
    GenericTypes = GenericAlias, type(t.List)


MigrationWriter.register_serializer(GenericContainer, GenericSerializer)
MigrationWriter.register_serializer(t.ForwardRef, TypingSerializer)
MigrationWriter.register_serializer(type(t.Union), TypingSerializer)  # type: ignore


if sys.version_info >= (3, 10):
    UnionType = types.UnionType

    class UnionTypeSerializer(BaseSerializer):
        value: UnionType

        def serialize(self):
            imports = set()
            if isinstance(self.value, type(t.Union)):  # type: ignore
                imports.add("import typing")

            for arg in get_args(self.value):
                _, arg_imports = serializer_factory(arg).serialize()
                imports.update(arg_imports)

            return repr(self.value), imports

    MigrationWriter.register_serializer(UnionType, UnionTypeSerializer)



import json
import typing as t
from functools import partial

import pydantic
from django.core import exceptions as django_exceptions
from django.db.models.fields import NOT_PROVIDED
from django.db.models.fields.json import JSONField
from django.db.models.query_utils import DeferredAttribute

from . import base, forms, utils
from ._migration_serializers import GenericContainer, GenericTypes

__all__ = ("SchemaField",)


class SchemaAttribute(DeferredAttribute):
    """
    Forces Django to call to_python on fields when setting them.
    This is useful when you want to add some custom field data postprocessing.

    Should be added to field like a so:

    ```
    def contribute_to_class(self, cls, name, *args, **kwargs):
        super().contribute_to_class(cls, name,  *args, **kwargs)
        setattr(cls, name, SchemaDeferredAttribute(self))
    ```
    """

    field: "PydanticSchemaField"

    def __set__(self, obj, value):
        obj.__dict__[self.field.attname] = self.field.to_python(value)


class PydanticSchemaField(JSONField, t.Generic[base.ST]):
    descriptor_class = SchemaAttribute
    is_prepared_schema: bool = False

    def __init__(
        self,
        *args,
        schema: t.Union[t.Type["base.ST"], "GenericContainer", "t.ForwardRef", str, None] = None,
        config: "base.ConfigType" = None,
        **kwargs,
    ):
        self.export_params = base.extract_export_kwargs(kwargs, dict.pop)
        super().__init__(*args, **kwargs)

        self.config = config
        self._resolve_schema(schema)

    def __copy__(self):
        _, _, args, kwargs = self.deconstruct()
        return type(self)(*args, **kwargs)

    def get_default(self):
        value = super().get_default()
        return self.to_python(value)

    def to_python(self, value) -> "base.SchemaT":
        # Attempt to resolve forward referencing schema if it was not succesful
        # during `.contribute_to_class` call
        if not self.is_prepared_schema:
            self._prepare_model_schema()
        try:
            assert self.decoder is not None
            return self.decoder().decode(value)
        except pydantic.ValidationError as e:
            raise django_exceptions.ValidationError(e.errors())

    def get_prep_value(self, value) -> t.Optional[str]:
        if not self.is_prepared_schema:
            self._prepare_model_schema()
        prep_value = super().get_prep_value(value)
        if prep_value is not None and not isinstance(prep_value, str):
            prep_value = self.encoder().encode(prep_value)  # type: ignore
        return prep_value

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        self._deconstruct_schema(kwargs)
        self._deconstruct_default(kwargs)
        self._deconstruct_config(kwargs)

        kwargs.pop("decoder")
        kwargs.pop("encoder")

        return name, path, args, kwargs

    def contribute_to_class(self, cls, name, private_only=False):
        if self.schema is None:
            self._resolve_schema_from_type_hints(cls, name)

        try:
            self._prepare_model_schema(cls)
        except NameError:
            # Pydantic was not able to resolve forward references, which means
            # that it should be postponed until initial access to the field
            self.is_prepared_schema = False

        super().contribute_to_class(cls, name, private_only)

    def formfield(self, **kwargs):
        if self.schema is None:
            self._resolve_schema_from_type_hints(self.model, self.attname)

        owner_model = getattr(self, "model", None)
        field_kwargs = dict(
            form_class=forms.SchemaField,
            schema=self.schema,
            config=self.config,
            __module__=getattr(owner_model, "__module__", None),
            **self.export_params,
        )
        field_kwargs.update(kwargs)
        return super().formfield(**field_kwargs)

    def _resolve_schema(self, schema):
        if isinstance(schema, GenericContainer):
            schema = t.cast(t.Type["base.ST"], schema.reconstruct_type())

        self.schema = schema
        if schema is not None:
            self.serializer_schema = serializer = base.wrap_schema(schema, self.config, self.null)
            self.decoder = partial(base.SchemaDecoder, serializer)  # type: ignore
            self.encoder = partial(base.SchemaEncoder, schema=serializer, export=self.export_params)  # type: ignore

    def _resolve_schema_from_type_hints(self, cls, name):
        annotated_schema = utils.get_annotated_type(cls, name)
        if annotated_schema is None:
            raise django_exceptions.FieldError(
                f"{cls._meta.label}.{name} needs to be either annotated "
                "or `schema=` field attribute should be explicitly passed"
            )
        self._resolve_schema(annotated_schema)

    def _prepare_model_schema(self, cls=None):
        cls = cls or getattr(self, "model", None)
        if cls is not None:
            base.prepare_schema(self.serializer_schema, cls)
            self.is_prepared_schema = True

    def _deconstruct_default(self, kwargs):
        default = kwargs.get("default", NOT_PROVIDED)

        if not (default is NOT_PROVIDED or callable(default)):
            # default value deconstruction with SchemaEncoder is meaningful
            # only if schema resolution is not deferred
            if self.is_prepared_schema:
                plain_default = self.get_prep_value(default)
            else:
                plain_default = default

            kwargs.update(default=plain_default)

    def _deconstruct_schema(self, kwargs):
        schema = self.schema
        if isinstance(schema, GenericTypes):
            schema = GenericContainer.from_generic(self.schema)

        kwargs.update(schema=schema)

    def _deconstruct_config(self, kwargs):
        kwargs.update(self.export_params, config=self.config)


if t.TYPE_CHECKING:
    OptSchemaT = t.Optional[base.SchemaT]


@t.overload
def SchemaField(
    schema: "t.Union[t.Type[t.Optional[base.ST]], t.ForwardRef]" = ...,
    config: "base.ConfigType" = ...,
    default: "t.Union[OptSchemaT, t.Callable[[], OptSchemaT]]" = ...,
    *args,
    null: "t.Literal[True]",
    **kwargs,
) -> "t.Optional[base.ST]":
    ...


@t.overload
def SchemaField(
    schema: "t.Union[t.Type[base.ST], t.ForwardRef]" = ...,
    config: "base.ConfigType" = ...,
    default: "t.Union[base.SchemaT, t.Callable[[], base.SchemaT]]" = ...,
    *args,
    null: "t.Literal[False]" = ...,
    **kwargs,
) -> "base.ST":
    ...


def SchemaField(schema=None, config=None, *args, **kwargs) -> t.Any:
    kwargs.update(schema=schema, config=config)
    return PydanticSchemaField(*args, **kwargs)



import sys
import typing as t
from datetime import date
from uuid import UUID

import pydantic
import pytest
from django_pydantic_field import base

from .conftest import InnerSchema, SampleDataclass


class SampleSchema(pydantic.BaseModel):
    __root__: InnerSchema


def test_schema_encoder():
    encoder = base.SchemaEncoder(schema=SampleSchema)
    existing_model_inst = InnerSchema(stub_str="abc", stub_list=[date(2022, 7, 1)])
    expected_encoded = '{"stub_str": "abc", "stub_int": 1, "stub_list": ["2022-07-01"]}'
    assert encoder.encode(existing_model_inst) == expected_encoded


def test_schema_encoder_with_raw_dict():
    encoder = base.SchemaEncoder(schema=SampleSchema)
    existing_raw = {"stub_str": "abc", "stub_list": [date(2022, 7, 1)]}
    expected_encoded = '{"stub_str": "abc", "stub_int": 1, "stub_list": ["2022-07-01"]}'
    assert encoder.encode(existing_raw) == expected_encoded


def test_schema_encoder_with_custom_config():
    encoder = base.SchemaEncoder(schema=SampleSchema, export={"exclude": {"__root__": {"stub_list"}}})
    existing_raw = {"stub_str": "abc", "stub_list": [date(2022, 7, 1)]}
    expected_encoded = '{"stub_str": "abc", "stub_int": 1}'
    assert encoder.encode(existing_raw) == expected_encoded


def test_schema_decoder():
    decoder = base.SchemaDecoder(schema=SampleSchema)
    existing_encoded = '{"stub_str": "abc", "stub_int": 1, "stub_list": ["2022-07-01"]}'
    expected_decoded = InnerSchema(stub_str="abc", stub_list=[date(2022, 7, 1)])

    assert decoder.decode(existing_encoded) == expected_decoded


def test_schema_decoder_error():
    existing_flawed_encoded = '{"stub_str": "abc", "stub_list": 1}'

    decoder = base.SchemaDecoder(schema=SampleSchema)

    with pytest.raises(pydantic.ValidationError) as e:
        decoder.decode(existing_flawed_encoded)

    assert e.match(".*stub_list.*")


def test_schema_wrapper_transformers():
    existing_encoded = '{"stub_str": "abc", "stub_int": 1, "stub_list": ["2022-07-01"]}'
    expected_decoded = InnerSchema(stub_str="abc", stub_list=[date(2022, 7, 1)])

    parsed_wrapper = base.wrap_schema(InnerSchema).parse_raw(existing_encoded)
    assert parsed_wrapper.__root__ == expected_decoded

    existing_encoded = '[{"stub_str": "abc", "stub_int": 1, "stub_list": ["2022-07-01"]}]'
    parsed_wrapper = base.wrap_schema(t.List[InnerSchema]).parse_raw(existing_encoded)
    assert parsed_wrapper.__root__ == [expected_decoded]


class test_schema_wrapper_config_inheritance():
    parsed_wrapper = base.wrap_schema(InnerSchema, config={"allow_mutation": False})
    assert not parsed_wrapper.Config.allow_mutation
    assert not parsed_wrapper.Config.frozen

    parsed_wrapper = base.wrap_schema(t.List[InnerSchema], config={"frozen": True})
    assert parsed_wrapper.Config.allow_mutation
    assert parsed_wrapper.Config.frozen


@pytest.mark.parametrize("type_, encoded, decoded", [
    (InnerSchema, '{"stub_str": "abc", "stub_list": ["2022-07-01"], "stub_int": 1}', InnerSchema(stub_str="abc", stub_list=[date(2022, 7, 1)])),
    (SampleDataclass, '{"stub_str": "abc", "stub_list": ["2022-07-01"], "stub_int": 1}', SampleDataclass(stub_str="abc", stub_list=[date(2022, 7, 1)])),
    (t.List[int], '[1, 2, 3]', [1, 2, 3]),
    (t.Mapping[int, date], '{"1": "1970-01-01"}', {1: date(1970, 1, 1)}),
    (t.Set[UUID], '["ba6eb330-4f7f-11eb-a2fb-67c34e9ac07c"]', {UUID("ba6eb330-4f7f-11eb-a2fb-67c34e9ac07c")}),
])
def test_concrete_types(type_, encoded, decoded):
    schema = base.wrap_schema(type_)
    encoder = base.SchemaEncoder(schema=schema)
    decoder = base.SchemaDecoder(schema=schema)

    existing_decoded = decoder.decode(encoded)
    assert existing_decoded == decoded

    existing_encoded = encoder.encode(decoded)
    assert decoder.decode(existing_encoded) == decoded


@pytest.mark.skipif(sys.version_info < (3, 9), reason="Should test against builtin generic types")
@pytest.mark.parametrize("type_factory, encoded, decoded", [
    (lambda: list[InnerSchema], '[{"stub_str": "abc", "stub_list": ["2022-07-01"], "stub_int": 1}]', [InnerSchema(stub_str="abc", stub_list=[date(2022, 7, 1)])]),
    (lambda: list[SampleDataclass], '[{"stub_str": "abc", "stub_list": ["2022-07-01"], "stub_int": 1}]', [SampleDataclass(stub_str="abc", stub_list=[date(2022, 7, 1)])]), # type: ignore
    (lambda: list[int], '[1, 2, 3]', [1, 2, 3]),
    (lambda: dict[int, date], '{"1": "1970-01-01"}', {1: date(1970, 1, 1)}),
    (lambda: set[UUID], '["ba6eb330-4f7f-11eb-a2fb-67c34e9ac07c"]', {UUID("ba6eb330-4f7f-11eb-a2fb-67c34e9ac07c")}),
])
def test_concrete_raw_types(type_factory, encoded, decoded):
    type_ = type_factory()

    schema = base.wrap_schema(type_)
    encoder = base.SchemaEncoder(schema=schema)
    decoder = base.SchemaDecoder(schema=schema)

    existing_decoded = decoder.decode(encoded)
    assert existing_decoded == decoded

    existing_encoded = encoder.encode(decoded)
    assert decoder.decode(existing_encoded) == decoded


@pytest.mark.parametrize("forward_ref, sample_data", [
    (t.ForwardRef("t.List[int]"), '[1, 2]'),
    (t.ForwardRef("InnerSchema"), '{"stub_str": "abc", "stub_int": 1, "stub_list": ["2022-07-01"]}'),
    (t.ForwardRef("PostponedSchema"), '{"stub_str": "abc", "stub_int": 1, "stub_list": ["2022-07-01"]}'),
])
def test_forward_refs_preparation(forward_ref, sample_data):
    schema = base.wrap_schema(forward_ref)
    base.prepare_schema(schema, test_forward_refs_preparation)
    assert schema.parse_raw(sample_data).json() == sample_data


class PostponedSchema(pydantic.BaseModel):
    __root__: InnerSchema



import typing as t
from functools import partial

import pydantic
from django.core.exceptions import ValidationError
from django.forms.fields import InvalidJSONInput, JSONField

from . import base

__all__ = ("SchemaField",)


class SchemaField(JSONField, t.Generic[base.ST]):
    def __init__(
        self,
        schema: t.Union[t.Type["base.ST"], t.ForwardRef],
        config: t.Optional["base.ConfigType"] = None,
        __module__: t.Optional[str] = None,
        **kwargs
    ):
        self.schema = base.wrap_schema(
            schema,
            config,
            allow_null=not kwargs.get("required", True),
            __module__=__module__,
        )
        export_params = base.extract_export_kwargs(kwargs, dict.pop)
        decoder = partial(base.SchemaDecoder, self.schema)
        encoder = partial(
            base.SchemaEncoder,
            schema=self.schema,
            export=export_params,
            raise_errors=True,
        )
        kwargs.update(encoder=encoder, decoder=decoder)
        super().__init__(**kwargs)

    def to_python(self, value):
        try:
            return super().to_python(value)
        except pydantic.ValidationError as e:
            raise ValidationError(e.errors(), code="invalid")

    def bound_data(self, data, initial):
        try:
            return super().bound_data(data, initial)
        except pydantic.ValidationError:
            return InvalidJSONInput(data)

    def get_bound_field(self, form, field_name):
        base.prepare_schema(self.schema, form)
        return super().get_bound_field(form, field_name)



import sys
import typing as t


def get_annotated_type(obj, field, default=None) -> t.Any:
    try:
        if isinstance(obj, type):
            annotations = obj.__dict__["__annotations__"]
        else:
            annotations = obj.__annotations__

        return annotations[field]
    except (AttributeError, KeyError):
        return default


def get_local_namespace(cls) -> t.Dict[str, t.Any]:
    try:
        module = cls.__module__
        return vars(sys.modules[module])
    except (KeyError, AttributeError):
        return {}



import os
from contextlib import contextmanager

from django.core.management import call_command

MIGRATIONS_DIR = "tests/sample_app/migrations/"


def test_makemigrations_not_failing():
    call_command("makemigrations", "sample_app", "--noinput", "--dry-run")


def test_makemigrations_no_duplicates(capfd):
    with clean_dir(MIGRATIONS_DIR):
        call_command("makemigrations", "sample_app", "--noinput")
        capfd.readouterr()

        call_command("makemigrations", "sample_app", "--noinput", "--dry-run")
        out, _ = capfd.readouterr()

    assert "No changes detected in app 'sample_app'" in out


@contextmanager
def clean_dir(path):
    initial_files = dir_files(path)

    try:
        yield
    finally:
        new_files = dir_files(path) - initial_files
        for f_path in new_files:
            os.remove(f_path)


def dir_files(path):
    files = {f.path for f in os.scandir(path) if f.is_file()}
    return files - {os.path.join(path, ".gitignore")}



import io
import typing as t
from datetime import date

import pytest
import yaml
from django.urls import path
from django_pydantic_field import rest_framework
from rest_framework import exceptions, generics, schemas, serializers, views
from rest_framework.decorators import api_view, parser_classes, renderer_classes, schema
from rest_framework.response import Response

from .conftest import InnerSchema


class SampleSerializer(serializers.Serializer):
    field = rest_framework.SchemaField(schema=t.List[InnerSchema])


def test_schema_field():
    field = rest_framework.SchemaField(InnerSchema)
    existing_instance = InnerSchema(stub_str="abc", stub_list=[date(2022, 7, 1)])
    expected_encoded = {
        "stub_str": "abc",
        "stub_int": 1,
        "stub_list": [date(2022, 7, 1)],
    }

    assert field.to_representation(existing_instance) == expected_encoded
    assert field.to_internal_value(expected_encoded) == existing_instance

    with pytest.raises(serializers.ValidationError):
        field.to_internal_value(None)

    with pytest.raises(serializers.ValidationError):
        field.to_internal_value("null")


def test_field_schema_with_custom_config():
    field = rest_framework.SchemaField(
        InnerSchema, allow_null=True, exclude={"stub_int"}
    )
    existing_instance = InnerSchema(stub_str="abc", stub_list=[date(2022, 7, 1)])
    expected_encoded = {"stub_str": "abc", "stub_list": [date(2022, 7, 1)]}

    assert field.to_representation(existing_instance) == expected_encoded
    assert field.to_internal_value(expected_encoded) == existing_instance
    assert field.to_internal_value(None) is None
    assert field.to_internal_value("null") is None


def test_serializer_marshalling_with_schema_field():
    existing_instance = {
        "field": [InnerSchema(stub_str="abc", stub_list=[date(2022, 7, 1)])]
    }
    expected_data = {
        "field": [{"stub_str": "abc", "stub_int": 1, "stub_list": [date(2022, 7, 1)]}]
    }

    serializer = SampleSerializer(instance=existing_instance)
    assert serializer.data == expected_data

    serializer = SampleSerializer(data=expected_data)
    serializer.is_valid(raise_exception=True)
    assert serializer.validated_data == existing_instance


@pytest.mark.parametrize("export_kwargs", [
    {"include": {"stub_str", "stub_int"}},
    {"exclude": {"stub_list"}},
    {"exclude_unset": True},
    {"exclude_defaults": True},
    {"exclude_none": True},
    {"by_alias": True},
])
def test_field_export_kwargs(export_kwargs):
    field = rest_framework.SchemaField(InnerSchema, **export_kwargs)
    assert field.to_representation(InnerSchema(stub_str="abc", stub_list=[date(2022, 7, 1)]))


def test_invalid_data_serialization():
    invalid_data = {"field": [{"stub_int": "abc", "stub_list": ["abc"]}]}
    serializer = SampleSerializer(data=invalid_data)

    with pytest.raises(exceptions.ValidationError) as e:
        serializer.is_valid(raise_exception=True)

    assert e.match(r".*stub_str.*stub_int.*stub_list.*")


def test_schema_renderer():
    renderer = rest_framework.SchemaRenderer()
    existing_instance = InnerSchema(stub_str="abc", stub_list=[date(2022, 7, 1)])
    expected_encoded = (
        b'{"stub_str": "abc", "stub_int": 1, "stub_list": ["2022-07-01"]}'
    )

    assert renderer.render(existing_instance) == expected_encoded


def test_typed_schema_renderer():
    renderer = rest_framework.SchemaRenderer[InnerSchema]()
    existing_data = {"stub_str": "abc", "stub_list": [date(2022, 7, 1)]}
    expected_encoded = (
        b'{"stub_str": "abc", "stub_int": 1, "stub_list": ["2022-07-01"]}'
    )

    assert renderer.render(existing_data) == expected_encoded


def test_schema_parser():
    parser = rest_framework.SchemaParser[InnerSchema]()
    existing_encoded = '{"stub_str": "abc", "stub_int": 1, "stub_list": ["2022-07-01"]}'
    expected_instance = InnerSchema(stub_str="abc", stub_list=[date(2022, 7, 1)])

    assert parser.parse(io.StringIO(existing_encoded)) == expected_instance


@api_view(["POST"])
@schema(rest_framework.AutoSchema())
@parser_classes([rest_framework.SchemaParser[InnerSchema]])
@renderer_classes([rest_framework.SchemaRenderer[t.List[InnerSchema]]])
def sample_view(request):
    assert isinstance(request.data, InnerSchema)
    return Response([request.data])


class ClassBasedViewWithSerializer(generics.RetrieveAPIView):
    serializer_class = SampleSerializer
    schema = rest_framework.AutoSchema()


class ClassBasedView(views.APIView):
    parser_classes = [rest_framework.SchemaParser[InnerSchema]]
    renderer_classes = [rest_framework.SchemaRenderer[t.List[InnerSchema]]]

    def post(self, request, *args, **kwargs):
        assert isinstance(request.data, InnerSchema)
        return Response([request.data])


class ClassBasedViewWithSchemaContext(ClassBasedView):
    parser_classes = [rest_framework.SchemaParser]
    renderer_classes = [rest_framework.SchemaRenderer]

    def get_renderer_context(self):
        ctx = super().get_renderer_context()
        return dict(ctx, render_schema=t.List[InnerSchema])

    def get_parser_context(self, http_request):
        ctx = super().get_parser_context(http_request)
        return dict(ctx, parser_schema=InnerSchema)


@pytest.mark.parametrize(
    "view",
    [
        sample_view,
        ClassBasedView.as_view(),
        ClassBasedViewWithSchemaContext.as_view(),
    ],
)
def test_end_to_end_api_view(view, request_factory):
    expected_instance = InnerSchema(stub_str="abc", stub_list=[date(2022, 7, 1)])
    existing_encoded = (
        b'{"stub_str": "abc", "stub_int": 1, "stub_list": ["2022-07-01"]}'
    )

    request = request_factory.post(
        "/", existing_encoded, content_type="application/json"
    )
    response = view(request)

    assert response.data == [expected_instance]
    assert response.data[0] is not expected_instance

    assert response.rendered_content == b"[%s]" % existing_encoded


def test_openapi_serializer_schema_generation(request_factory):
    schema_url_patterns = [
        path("api/", ClassBasedViewWithSerializer.as_view()),
    ]

    schema_view = schemas.get_schema_view(patterns=schema_url_patterns)
    request = request_factory.get("api/", format="json")
    response = schema_view(request)

    results = yaml.load(response.rendered_content, yaml.CLoader)
    assert results["components"]["schemas"]["Sample"]["properties"]["field"] == {
        "title": "FieldSchema[List[tests.conftest.InnerSchema]]",
        "type": "array",
        "items": {"$ref": "#/definitions/InnerSchema"},
        "definitions": {
            "InnerSchema": {
                "title": "InnerSchema",
                "type": "object",
                "properties": {
                    "stub_str": {"title": "Stub Str", "type": "string"},
                    "stub_int": {
                        "title": "Stub Int",
                        "default": 1,
                        "type": "integer",
                    },
                    "stub_list": {
                        "title": "Stub List",
                        "type": "array",
                        "items": {"type": "string", "format": "date"},
                    },
                },
                "required": ["stub_str", "stub_list"],
            }
        },
    }


def test_openapi_parser_renderer_schema_generation(request_factory):
    schema_url_patterns = [
        path("api/", sample_view),
    ]

    schema_view = schemas.get_schema_view(patterns=schema_url_patterns)
    request = request_factory.get("api/", format="json")
    response = schema_view(request)

    results = yaml.load(response.rendered_content, yaml.CLoader)
    assert results["paths"]["/api/"]["post"]["requestBody"]["content"][
        "application/json"
    ] == {
        "schema": {
            "title": "FieldSchema[InnerSchema]",
            "$ref": "#/definitions/InnerSchema",
            "definitions": {
                "InnerSchema": {
                    "title": "InnerSchema",
                    "type": "object",
                    "properties": {
                        "stub_str": {
                            "title": "Stub Str",
                            "type": "string",
                        },
                        "stub_int": {
                            "title": "Stub Int",
                            "default": 1,
                            "type": "integer",
                        },
                        "stub_list": {
                            "title": "Stub List",
                            "type": "array",
                            "items": {
                                "type": "string",
                                "format": "date",
                            },
                        },
                    },
                    "required": ["stub_str", "stub_list"],
                }
            },
        }
    }
    assert results["paths"]["/api/"]["post"]["responses"]["201"] == {
        "content": {
            "application/json": {
                "schema": {
                    "title": "FieldSchema[List[tests.conftest.InnerSchema]]",
                    "type": "array",
                    "items": {"$ref": "#/definitions/InnerSchema"},
                    "definitions": {
                        "InnerSchema": {
                            "title": "InnerSchema",
                            "type": "object",
                            "properties": {
                                "stub_str": {
                                    "title": "Stub Str",
                                    "type": "string",
                                },
                                "stub_int": {
                                    "title": "Stub Int",
                                    "default": 1,
                                    "type": "integer",
                                },
                                "stub_list": {
                                    "title": "Stub List",
                                    "type": "array",
                                    "items": {
                                        "type": "string",
                                        "format": "date",
                                    },
                                },
                            },
                            "required": ["stub_str", "stub_list"],
                        }
                    },
                }
            }
        },
        "description": "",
    }
